var set_names = ["base", "intrigue", "seaside", "alchemy"];

var spot_to_change = 0;
var spot_options = null;

var available_cards = [];

var chosen_cards = [];
var black_market_deck = null;
var black_market_options = [];

var last_chosen_cards = null;
var last_black_market_deck = null;
var last_black_market_options = null;

var required_cards = [];

function start() {
    spot_to_change = 0;
    spot_options = null;
    names = ["cellar", "market", "militia", "mine", "moat",
	     "remodel", "smithy", "village", "woodcutter", "workshop"];
    create_all_cards();
    update_available_cards();
    populate_cards_table("promo_cards_table", [envoy_card, black_market_card, stash_card]);
    if (document.getElementById("require_black_market").checked)
	require_card("black_market");
    chosen_cards = get_cards_by_names(names);
    init_black_market_deck();
    show_chosen_cards();
    show_black_market_options();
    show_or_hide_constraints();
    show_or_hide_card_chooser();
    show_or_hide_all_cards();
    clear_undo();
}

function populate_cards_table(table_id, cards) {
    var table = document.getElementById(table_id);

    for (var i = table.rows.length - 1; i >= 0; i--)
	table.deleteRow(i);

    var i = 0;
    for (var y = 0; ; y++) {
	var row = table.insertRow(-1);
	for (var x = 0; x < 5; x++) {
	    var card = cards[i++];
	    var col = row.insertCell(-1);
	    var img = document.createElement("img");

	    img.id = card.name + "_card";
	    img.alt = "[" + card.name + "]";
	    img.src = "images/" + card.name + ".jpg";
	    img.width = 185;
	    img.height = 285;
	    img.className = "not_selected";
	    img.setAttribute("onclick", "toggle_card('" + card.name + "');");
	    col.appendChild(img);
	    if (i >= cards.length)
		return;
	}
    }
}

function require_card(name) {
    var id = document.getElementById(name + "_card");
    var card = find_card_by_name(name, all_cards);
    id.className = card.set;
    required_cards.push(card);
    if (name == "black_market")
	document.getElementById("require_black_market").checked = true;
}

function unrequire_card(name) {
    var id = document.getElementById(name + "_card");
    var card = find_card_by_name(name, required_cards);
    id.className = "not_selected";
    remove_items([card], required_cards);
    if (name == "black_market")
	document.getElementById("require_black_market").checked = false;
}

function toggle_card(name) {
    var id = document.getElementById(name + "_card");
    if (id.className == "not_selected")
	require_card(name);
    else
	unrequire_card(name);
}

function parse_numeric_input_field(id) {
    return parseInt(document.getElementById(id).value);
}

/*
 * Returns an associative array object, indexed by card set name,
 * which specifies the minimum number of cards from that set
 * (including the individually-specified required cards).
 */
function get_min_cards() {
    var min_cards = new Object;

    min_cards["promo"] = 0;
    for (var i = 0; i < set_names.length; i++) {
	var num = parse_numeric_input_field("min_cards_" + set_names[i]);
	if (num < 0 || isNaN(num))
	    num = 0;
	min_cards[set_names[i]] = num;
    }

    return min_cards;
}

/*
 * Returns an associative array object, indexed by card set name,
 * which specifies the minimum number of cards from that set
 * (not including the individually-specified required cards).
 */
function get_min_cards_except_required_cards() {
    var min_cards = get_min_cards();

    for (var i = 0; i < required_cards.length; i++)
	if (min_cards[required_cards[i].set] > 0)
	    min_cards[required_cards[i].set]--;

    return min_cards;
}

function requirements_obviously_unsatisfiable() {
    if (available_cards.length < 10) {
	alert("There are fewer than 10 cards available.");
	return true;
    }

    if (required_cards.length > 10) {
	alert("You've specified more than 10 required cards.");
	return true;
    }

    var min_cards = get_min_cards_except_required_cards();
    var spots_left = 10 - required_cards.length;
    for (var i = 0; i < set_names.length; i++)
	spots_left -= min_cards[set_names[i]];
    if (spots_left < 0) {
	alert("The combination of required cards and per-card-set minimums adds up to more than 10 required cards.");
	return true;
    }
    
    return false;
}

function choose_new_cards() {
    if (requirements_obviously_unsatisfiable())
	return;
    else {
	var min_cards_by_set = get_min_cards_except_required_cards();
	var max_attempts = 200;

	var options_by_set = new Object;
	for (var i = 0; i < set_names.length; i++) {
	    var options = cards_by_set[set_names[i]].slice(0);
	    remove_items(required_cards, options);
	    options_by_set[set_names[i]] = options;
	}

	for (var i = 0; i < max_attempts; i++) {
	    var cards = required_cards.slice(0);
	    for (var i = 0; i < set_names.length; i++)
		cards = cards.concat(choose_cards_randomly(min_cards_by_set[set_names[i]], options_by_set[set_names[i]]));
	    var remaining_options = available_cards.slice(0);
	    remove_items(cards, remaining_options);
	    cards = cards.concat(choose_cards_randomly(10 - cards.length, remaining_options));
	    if (card_set_satisfies_constraints(cards))
		break;
	}
	if (i == max_attempts)
	    alert ("Unable to satisfy the constraints.");
	else {
	    save_state_for_undo();
	    spot_to_change = 0;
	    spot_options = null;
	    chosen_cards = cards;
	    init_black_market_deck();
	    show_chosen_cards();
	    show_black_market_options();
	}
    }
}

function replace_card(spot) {
    if (spot_to_change != spot) {
	spot_to_change = spot;
	spot_options = null;
    }
    if (available_cards.length < 11)
	alert("There are fewer than 11 cards available.");
    else if (find_card_by_name(chosen_cards[spot].name, required_cards))
	alert("You said that was a required card, so it can't be replaced.");
    else {
	var cards = chosen_cards.slice(0);
	var orig_card = cards[spot];

	if (spot_options === null) {
	    spot_options = choose_cards_randomly(available_cards.length, available_cards);
	    remove_items(cards, spot_options);
	}
	for (var i = 0; i < spot_options.length; i++) {
	    cards.splice(spot, 1, spot_options[i]);
	    if (card_set_satisfies_constraints(cards))
		break;
	}
	if (i == spot_options.length)
	    alert ("Unable to replace that card without violating the constraints.");
	else {
	    save_state_for_undo();
	    var last_spot_options = spot_options;
	    spot_options = last_spot_options.slice(i+1);
	    spot_options.push(orig_card);
	    spot_options = spot_options.concat(last_spot_options.slice(0, i));
	    chosen_cards = cards;
	    init_black_market_deck();
	    show_chosen_cards();
	    show_black_market_options();
	}
    }
}

function reset_game() {
    save_state_for_undo();
    init_black_market_deck();
    show_black_market_options();
    alert("The black market deck has been reset.");
}

function init_black_market_deck() {
    black_market_deck = null;
    black_market_options = [];
}

function black_market() {
    if (!black_market_is_there())
	alert("Black Market is not in the current set of cards.");
    else if (black_market_options.length > 0)
	alert("You must first choose a card from your last Black Market.");
    else {
	if (black_market_deck === null) {
	    black_market_deck = choose_cards_randomly(available_cards.length, available_cards);
	    remove_items(chosen_cards, black_market_deck);
	}
	if (black_market_deck.length == 0)
	    alert("No cards left in the Black Market deck.");
	else {
	    document.getElementById("reset_game_button").disabled = false;
	    clear_undo();
	    if (black_market_deck.length < 3) {
		black_market_options = black_market_deck;
		black_market_deck = [];
	    } else
		black_market_options = black_market_deck.splice(0, 3);
	    show_black_market_options();
	}
    }
}

function black_market_choose_none() {
    if (black_market_options.length > 0) {
	save_state_for_undo();
	black_market_deck = black_market_deck.concat(black_market_options);
	black_market_options = [];
	show_black_market_options();
    }
}

function black_market_choose(i) {
    if (i < black_market_options.length) {
	save_state_for_undo();
	black_market_options.splice(i, 1);
	black_market_deck = black_market_deck.concat(black_market_options);
	black_market_options = [];
	show_black_market_options();
    }
}

function black_market_is_there() {
    if (find_card_by_name("black_market", chosen_cards))
	return true;
    else
	return false;
}

function show_black_market_options() {
    var cards = black_market_options;
    for (var i = 0; i < 3; i++) {
	var card = (i < cards.length) ? cards[i] : null;
	var img = document.getElementById("bm" + i);
	show_card(card, img);
    }
    document.getElementById("black_market_helper").className = black_market_is_there() ? "black_market_helper" : "hidden";
    document.getElementById("black_market_button").disabled = (((black_market_deck !== null) &&
								(black_market_deck.length == 0)) ||
							       (black_market_options.length > 0));
    document.getElementById("choose_none_button").disabled = (black_market_options.length == 0);
    document.getElementById("reset_game_button").disabled = (black_market_deck === null);
}

function show_chosen_cards() {
    var cards = chosen_cards;
    for (var i = 0; i < 10; i++) {
	var card = (i < cards.length) ? cards[i] : null;
	var img = document.getElementById("card" + i);
	show_card(card, img);
    }
}

function show_card(card, img) {
    if (card === null) {
	img.alt = "[]";
	img.src = "images/blank.png";
	img.className = "not_selected";
    } else {
	img.alt = "[" + card.name + "]";
	img.src = "images/" + card.name + ".jpg";
	img.className = card.set;
    }
}

function find_card_by_name(name, cards) {
    for (var i = 0; i < cards.length; i++)
	if (cards[i].name == name)
	    return cards[i];
    return null;
}

function get_cards_by_names(names) {
    var cards = [];
    
    for (var i = 0; i < names.length; i++)
	cards.push(find_card_by_name(names[i], all_cards));
    return cards;
}

function remove_items(items, lst) {
    for (var i = 0; i < items.length; i++)
	for (var j = 0; j < lst.length; j++) {
	    if (lst[j] === items[i]) {
		lst.splice(j, 1);
		break;
	    }
	}
}

function choose_cards_randomly(n, cards) {
    var remaining = cards.slice(0);
    var chosen = [];
    if (n <= cards.length)
	for (var i = 0; i < n; i++) {
	    var x = random_int(remaining.length);
	    chosen.push(remaining[x]);
	    remaining.splice(x, 1);
	}
    return chosen;
}

function random_int(n) {
    var i;
    do {
	i = Math.floor(Math.random() * n);
    } while ((i < 0) || (i >= n));
    return i;
}

function clear_undo() {
    last_chosen_cards = null;
    last_black_market_deck = null;
    last_black_market_options = null;
    document.getElementById("undo_button_1").disabled = true;
    document.getElementById("undo_button_2").disabled = true;
}

function save_state_for_undo() {
    last_chosen_cards = chosen_cards.slice(0);
    if (black_market_deck === null)
	last_black_market_deck = null;
    else
	last_black_market_deck = black_market_deck.slice(0);
    last_black_market_options = black_market_options.slice(0);
    document.getElementById("undo_button_1").disabled = false;
    document.getElementById("undo_button_2").disabled = false;
}

function undo_last_action() {
    if (last_chosen_cards === null)
	alert("You cannot undo the last action.");
    else {
	var tmp;

	tmp = chosen_cards;
	chosen_cards = last_chosen_cards;
	last_chosen_cards = tmp;

	tmp = black_market_deck;
	black_market_deck = last_black_market_deck;
	last_black_market_deck = tmp;

	tmp = black_market_options;
	black_market_options = last_black_market_options;
	last_black_market_options = tmp;

	spot_options = null;

	show_chosen_cards();
	show_black_market_options();
    }
}

function show_or_hide(id) {
    if (document.getElementById("show_" + id).checked)
	document.getElementById(id).className = id;
    else
	document.getElementById(id).className = "hidden";
}

function show_or_hide_card_chooser() {
    show_or_hide("card_chooser");
}

function show_or_hide_constraints() {
    show_or_hide("constraints");
}

function show_or_hide_all_cards() {
    if (document.getElementById("show_all_cards").checked &&
	    !document.getElementById("adventurer_card")) {
	for (var i = 0; i < set_names.length; i++)
	    populate_cards_table(set_names[i] + "_cards_table", cards_by_set[set_names[i]]);
    }
    show_or_hide("all_cards");
}

function update_available_cards() {
    available_cards = [];
    spot_options = null;
    if (document.getElementById("include_black_market").checked)
	available_cards.push(black_market_card);
    if (document.getElementById("include_envoy").checked)
	available_cards.push(envoy_card);
    if (document.getElementById("include_stash").checked)
	available_cards.push(stash_card);
    for (var i = 0; i < set_names.length; i++)
	if (document.getElementById("include_" + set_names[i]).checked)
	    available_cards = available_cards.concat(cards_by_set[set_names[i]]);
}

function create_all_cards() {
    all_cards = [];
    all_cards.push(black_market_card);
    all_cards.push(envoy_card);
    all_cards.push(stash_card);
    for (var i = 0; i < set_names.length; i++)
	all_cards = all_cards.concat(cards_by_set[set_names[i]]);
}

function card_set_satisfies_constraints(cards) {
    var has_add_1_buy = false;
    var has_add_2_actions = false;
    var has_attack = false;
    var has_defense = false;
    var count_by_cost = [0,0,0,0,0,0,0,0,0];
    var missing_required_cards = required_cards.slice(0);
    var missing_count_by_set = get_min_cards();
    
    for (var i = 0; i < 10; i++) {
	remove_items([cards[i]], missing_required_cards);
	count_by_cost[cards[i].cost]++;
	missing_count_by_set[cards[i].set]--;
	if (cards[i].addbuys)
	    has_add_1_buy = true;
	if (cards[i].addactions && (cards[i].addactions >= 2))
	    has_add_2_actions = true;
	if (cards[i].attack)
	    has_attack = true;
	if (cards[i].reaction || cards[i].defense)
	    has_defense = true;
    }
    if (document.getElementById("require_add_1_buy").checked && !has_add_1_buy)
	return false;
    if (document.getElementById("require_add_2_actions").checked && !has_add_2_actions)
	return false;
    if (document.getElementById("if_attack_require_defense").checked && has_attack && !has_defense)
	return false;
    if (missing_required_cards.length > 0)
	return false;
    for (var i = 0; i < set_names.length; i++)
	if (missing_count_by_set[set_names[i]] > 0)
	    return false;
    if (document.getElementById("require_equitable_costs").checked)
	for (var i = 0; i < count_by_cost.length; i++)
	    if ((count_by_cost[i] > 3) || ((count_by_cost[i] < 1) && (i >= 2) && (i <= 5)))
		return false;
    return true;
}

//
// The card attributes are:
//
//   name     (string) [required]
//   set      (string) [required]
//   cost     (number) [required]
//
// (the following booleans need only be specified when they are true)
//
//   action      (boolean) (means this is an action card)
//   victory     (boolean) (means this is a victory card)
//   attack      (boolean) (means this is an attack card)
//   reaction    (boolean) (means this is a reaction card) [automatically implies defense]
//   defense     (boolean) (means this card is useful for defense)
//   duration    (boolean) (means this is a duration card)
//   interactive (boolean) (means this card involves interaction with other players somehow)
//   trashthis   (boolean) (means this card can be used to trash itself)
//   trashme     (boolean or "cost") (means this card can be used to trash my cards ["cost" means it adds coins equal to cost of trashed card])
//   curse       (boolean or "maybe") (means this card can give curses)
//
//   coins      (number) (means this card can be used as money (of the specified value), without using an action)
//   addcoins   (number) (how many additional coins this card gives, when used as an action)
//   addcards   (number) (how many additional cards this card gives, when used as an action)
//   addactions (number) (how many additional actions this card gives, when used as an action)
//   addbuys    (number) (how many additional buys this card gives, when used as an action)
//   victorypts (number) (how many victory points this card is worth)
//   gaincard (number or string)  (means this card can be used to gain a card of the given cost [if a number], or a particular card [if a string], e.g. "silver")
//   resupply (boolean) (means this card can be used to put cards back on the supply pile)

var black_market_card = { name: "black_market", set: "promo", action: true, cost: 3, addcoins: 2 };
var envoy_card = { name: "envoy", set: "promo", action: true, cost: 4, interactive: true };
var stash_card = { name: "stash", set: "promo", cost: 5, coins: 2 };

var cards_by_set = new Object;

cards_by_set["base"] = [
  { name: "adventurer", set: "base", action: true, cost: 6 }, 
  { name: "bureaucrat", set: "base", action: true, attack: true, cost: 4 }, 
  { name: "cellar", set: "base", action: true, cost: 2, addactions: 1 }, 
  { name: "chancellor", set: "base", action: true, cost: 3, addcoins: 2 }, 
  { name: "chapel", set: "base", action: true, cost: 2, trashme: true }, 
  { name: "council_room", set: "base", action: true, cost: 5, addcards: 4, addbuys: 1, interactive: true }, 
  { name: "feast", set: "base", action: true, cost: 4, trashme: true, gaincard: 5 }, 
  { name: "festival", set: "base", action: true, cost: 5, addactions: 2, addbuys: 1, addcoins: 2 }, 
  { name: "gardens", set: "base", victory: true, cost: 4 }, 
  { name: "laboratory", set: "base", action: true, cost: 5, addcards: 2, addactions: 1 }, 
  { name: "library", set: "base", action: true, cost: 5 }, 
  { name: "market", set: "base", action: true, cost: 5, addcards: 1, addactions: 1, addbuys: 1, addcoins: 1 }, 
  { name: "militia", set: "base", action: true, attack: true, cost: 4, addcoins: 2 }, 
  { name: "mine", set: "base", action: true, cost: 5, trashme: true }, 
  { name: "moat", set: "base", action: true, reaction: true, cost: 2, addcards: 2 }, 
  { name: "moneylender", set: "base", action: true, cost: 4 }, 
  { name: "remodel", set: "base", action: true, cost: 4, trashme: 2 }, 
  { name: "smithy", set: "base", action: true, cost: 4, addcards: 3 }, 
  { name: "spy", set: "base", action: true, attack: true, cost: 4, addcards: 1, addactions: 1 },
  { name: "thief", set: "base", action: true, attack: true, cost: 4 }, 
  { name: "throne_room", set: "base", action: true, cost: 4 }, 
  { name: "village", set: "base", action: true, cost: 3, addcards: 1, addactions: 2 }, 
  { name: "witch", set: "base", action: true, attack: true, cost: 5, addcards: 2, curse: true }, 
  { name: "woodcutter", set: "base", action: true, cost: 3, addbuys: 1, addcoins: 2 }, 
  { name: "workshop", set: "base", action: true, cost: 3, gaincard: 4 } ];

cards_by_set["intrigue"] = [
  { name: "baron", set: "intrigue", action: true, cost: 4, addbuys: 1 }, 
  { name: "bridge", set: "intrigue", action: true, cost: 4, addbuys: 1, addcoins: 1 }, 
  { name: "conspirator", set: "intrigue", action: true, cost: 4, addcoins: 2, addcards: 1, addactions: 1 }, 
  { name: "coppersmith", set: "intrigue", action: true, cost: 4 },
  { name: "courtyard", set: "intrigue", action: true, cost: 2, addcards: 3 }, 
  { name: "duke", set: "intrigue", victory: true, cost: 5 }, 
  { name: "great_hall", set: "intrigue", action: true, victory: true, cost: 3, addcards: 1, addactions: 1, victorypts: 1 }, 
  { name: "harem", set: "intrigue", treasure: true, victory: true, cost: 6, victorypts: 2, coins: 2 }, 
  { name: "ironworks", set: "intrigue", action: true, cost: 4, gaincard: 4, addactions: 1, addcoins: 1, addcards: 1 }, 
  { name: "masquerade", set: "intrigue", action: true, cost: 3, addcards: 2, interactive: true }, 
  { name: "mining_village", set: "intrigue", action: true, cost: 4, addcards: 1, addactions: 2, trashme: true, addcoins: 2 }, 
  { name: "minion", set: "intrigue", action: true, attack: true, cost: 5, addactions: 1, addcoins: 2 }, 
  { name: "nobles", set: "intrigue", action: true, victory: true, cost: 6, addcards: 3, addactions: 2, victorypts: 2 }, 
  { name: "pawn", set: "intrigue", action: true, cost: 2, addcards: 1, addactions: 1, addbuys: 1, addcoins: 1 }, 
  { name: "saboteur", set: "intrigue", action: true, attack: true, cost: 5 }, 
  { name: "scout", set: "intrigue", action: true, cost: 4, addactions: 1 }, 
  { name: "secret_chamber", set: "intrigue", action: true, reaction: true, cost: 2 }, 
  { name: "shanty_town", set: "intrigue", action: true, cost: 3, addactions: 2, addcards: 2 }, 
  { name: "steward", set: "intrigue", action: true, cost: 3, addcards: 2, addcoins: 2, trashme: true }, 
  { name: "swindler", set: "intrigue", action: true, attack: true, cost: 3, addcoins: 2 }, 
  { name: "torturer", set: "intrigue", action: true, attack: true, cost: 5, addcards: 3, curse: "maybe" }, 
  { name: "trading_post", set: "intrigue", action: true, cost: 5, trashme: true }, 
  { name: "tribute", set: "intrigue", action: true, cost: 5, addactions: 2, addcoins: 2, addcards: 2, interactive: true }, 
  { name: "upgrade", set: "intrigue", action: true, cost: 5, addcards: 1, addactions: 1, trashme: 1 }, 
  { name: "wishing_well", set: "intrigue", action: true, cost: 3, addcards: 1, addactions: 1 } ];

cards_by_set["seaside"] = [
  { name: "ambassador", set: "seaside", action: true, attack: true, cost: 3, resupply: true }, 
  { name: "bazaar", set: "seaside", action: true, cost: 5, addcards: 1, addactions: 2, addcoins: 1 }, 
  { name: "caravan", set: "seaside", action: true, duration: true, cost: 4, addcards: 1, addactions: 1 }, 
  { name: "cutpurse", set: "seaside", action: true, attack: true, cost: 4, addcoins: 2 }, 
  { name: "embargo", set: "seaside", action: true, cost: 2, addcoins: 2, trashthis: true }, 
  { name: "explorer", set: "seaside", action: true, cost: 5, gaincard: "silver" }, 
  { name: "fishing_village", set: "seaside", action: true, duration: true, cost: 3, addactions: 2, addcoins: 1 }, 
  { name: "ghost_ship", set: "seaside", action: true, attack: true, cost: 5, addcards: 2 }, 
  { name: "haven", set: "seaside", action: true, duration: true, cost: 2, addcards: 1, addactions: 1 }, 
  { name: "island", set: "seaside", action: true, victory: true, cost: 4, victorypts: 2 }, 
  { name: "lighthouse", set: "seaside", action: true, duration: true, cost: 2, defense: true, addactions: 1 },
  { name: "lookout", set: "seaside", action: true, cost: 3, addactions: 1, trashme: true }, 
  { name: "merchant_ship", set: "seaside", action: true, duration: true, cost: 5, addcoins: 2 }, 
  { name: "native_village", set: "seaside", action: true, cost: 2, addactions: 2 }, 
  { name: "navigator", set: "seaside", action: true, cost: 4, addcoins: 2 }, 
  { name: "outpost", set: "seaside", action: true, duration: true, cost: 5 }, 
  { name: "pearl_diver", set: "seaside", action: true, cost: 2, addcards: 1, addactions: 1 }, 
  { name: "pirate_ship", set: "seaside", action: true, attack: true, cost: 4 }, 
  { name: "salvager", set: "seaside", action: true, cost: 4, addbuys: 1, trashme: "cost" }, 
  { name: "sea_hag", set: "seaside", action: true, attack: true, cost: 4, curse: true }, 
  { name: "smugglers", set: "seaside", action: true, cost: 3 }, 
  { name: "tactician", set: "seaside", action: true, duration: true, cost: 5, addcards: 5, addbuys: 1, addactions: 1 }, 
  { name: "treasure_map", set: "seaside", action: true, cost: 4, trashthis: true }, 
  { name: "treasury", set: "seaside", action: true, cost: 5, addcards: 1, addactions: 1, addcoins: 1 }, 
  { name: "warehouse", set: "seaside", action: true, cost: 3, addcards: 3, addactions: 1 }, 
  { name: "wharf", set: "seaside", action: true, duration: true, cost: 5, addcards: 2, addbuys: 1 } ];

cards_by_set["alchemy"] = [
  { name: "alchemist", set: "alchemy", action: true, cost: 3, costspotion: true, addcards: 2, addactions: 1 }, 
  { name: "apothecary", set: "alchemy", action: true, cost: 2, costspotion: true, addcards: 1, addactions: 1 }, 
  { name: "apprentice", set: "alchemy", action: true, cost: 5, costspotion: true, addactions: 1, trashme: true }, 
  { name: "familiar", set: "alchemy", action: true, attack: true, cost: 3, costspotion: true, addcards: 1, addactions: 1, curse: true }, 
  { name: "golem", set: "alchemy", action: true, cost: 4, costspotion: true }, 
  { name: "herbalist", set: "alchemy", action: true, cost: 2, addbuys: 1, addcoins: 1 },
  { name: "philosophers_stone", set: "alchemy", cost: 3, costspotion: true }, 
  { name: "possession", set: "alchemy", action: true, cost: 6, costspotion: true, interactive: true }, 
  { name: "scrying_pool", set: "alchemy", action: true, attack: true, cost: 2, costspotion: true, addactions: 1, interactive: true }, 
  { name: "transmute", set: "alchemy", action: true, cost: 0, costspotion: true, trashme: true }, 
  { name: "university", set: "alchemy", action: true, cost: 2, costspotion: true, addactions: 2, gaincard: 5 }, 
  { name: "vineyard", set: "alchemy", victory: true, cost: 0, costspotion: true } ];
